-------------------------------------------------------------------------------
-- Localized Lua globals.
-------------------------------------------------------------------------------
local _G = getfenv(0)

local string = _G.string

local table = _G.table
local tremove, tinsert, tsort = table.remove, table.insert, table.sort

local ipairs, pairs = _G.ipairs, _G.pairs
local wipe = _G.wipe

local math = _G.math
local floor = math.floor

local tostring = _G.tostring

local GetTime = _G.GetTime

-------------------------------------------------------------------------------
-- Addon namespace
-------------------------------------------------------------------------------
local nanoTalk = LibStub("AceAddon-3.0"):NewAddon("nanoTalk", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")

local DataObj = LDB:NewDataObject("nanoTalk",
	{
		type	= "data source",
		icon	= "Interface\\GossipFrame\\GossipGossipIcon",
		label	= "nanoTalk",
		text	= "0t",
	})

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
local COLUMNS = 3
local VERSION = GetAddOnMetadata("nanoTalk", "Version")

local defaults = {
	profile = {
		ChatNotify = true,
		NotifyAFK = true,
		SoundAlert = true,
		TalkTime = 2,
		TipScale = 1.0,
		TrackNum = 10,
		TrackTime = 120,
		Version = VERSION
	},
	factionrealm = {
		History = {}
	},
}

-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local trackSecs, talkSecs
local timers = {}
local updatedLines = {}
local timestamps = {}
local people = {}
local AFKtime = 0

local tooltip
local db

-------------------------------------------------------------------------------
-- Local functions
-------------------------------------------------------------------------------
local function SetTimes()
	trackSecs = (db.TrackTime * 60)
	talkSecs = (db.TalkTime * 60)
end

local function GetCount(timer)
	if not timer or timer == 0 then
		return #timestamps
	end
	local count = 0
	local now = GetTime()

	for _, timestamp in ipairs(timestamps) do
		if (now - timestamp) <= timer then
			count = count + 1
		end
	end
	return count
end

local function TimeStr(seconds)
	local tm = seconds
	local timeval = tostring(floor(tm % 60)).."s"

	if tm >= 60 then
		tm = tm / 60
		timeval = tostring(floor(tm % 60)).."m"..timeval

		if tm >= 60 then
			tm = tm / 60
			timeval = tostring(floor(tm % 24)).."h"..timeval

			if tm >= 24 then
				tm = tm / 24
				timeval = tostring(floor(tm)).."d"..timeval
			end
		end
	end
	return timeval
end

local function ShowTooltip()
	tooltip:Clear()
	tooltip:AddHeader(" ")
	tooltip:SetCell(1, 1, "nanoTalk", "CENTER", COLUMNS)
	tooltip:AddLine(" ")

	local now = GetTime()
	local count = GetCount(trackSecs)
	local line

	if count > 0 then
		tooltip:AddHeader("|cffff9933Name|r", "|cffff9933From|r", "|cffff9933Last|r")

		for _, tm in ipairs(timestamps) do
			local cur_time = now - tm

			if trackSecs == 0 or cur_time <= trackSecs then
				line = tooltip:AddLine(people[tm].name, people[tm].msgs, TimeStr(cur_time))
				updatedLines[line] = tm
			end
		end
		tooltip:AddLine(" ")

		line = tooltip:AddLine(" ")
		tooltip:SetCell(line, 1, "CTRL+Left Click: Clear list.", "CENTER", COLUMNS)
	else
		tooltip:AddLine("No messages.")
		tooltip:AddLine(" ")
	end
	line = tooltip:AddLine(" ")
	tooltip:SetCell(line, 1, "Right Click: Configure", "CENTER", COLUMNS)
	tooltip:Show()
end

local function UpdateTooltip()
	if not tooltip or not tooltip:IsShown() then
		return
	end
	wipe(updatedLines)
	ShowTooltip()
end

-------------------------------------------------------------------------------
-- LDB functions
-------------------------------------------------------------------------------
local QTip = LibStub:GetLibrary("LibQTip-1.0")

local function HideTooltip()
	if not tooltip or tooltip:IsMouseOver() then
		return
	end
	tooltip:SetScript("OnLeave", nil)
	tooltip = QTip:Release(tooltip)
	wipe(updatedLines)
end

function DataObj.OnLeave(display, motion)
	HideTooltip()
end

function DataObj:OnClick(button)
	if button == "LeftButton" and IsControlKeyDown() then
		nanoTalk:Clear()
	else
		InterfaceOptionsFrame_OpenToCategory(nanoTalk.optionsFrames.nanoTalk)
	end
end

function DataObj.OnEnter(display, motion)
	tooltip = QTip:Acquire("nanoTalk", COLUMNS, "LEFT", "CENTER", "RIGHT")
	tooltip:SmartAnchorTo(display)
	tooltip:SetScale(db.TipScale)
	tooltip:SetScript("OnLeave", HideTooltip)
	ShowTooltip()
end

-------------------------------------------------------------------------------
-- Initialization and event handlers
-------------------------------------------------------------------------------
function nanoTalk:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("nanoTalkConfig", defaults)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	db = self.db.profile

	SetTimes()
	self:SetupOptions()
end

function nanoTalk:OnEnable()
	self:RegisterEvent("CHAT_MSG_WHISPER")
	timers.UpdateInfo = self:ScheduleRepeatingTimer("UpdateInfo", 0.5)
end

function nanoTalk:OnDisable()
	self:UnregisterEvent("CHAT_MSG_WHISPER")

	for k, v in pairs(timers) do
		self:CancelTimer(v, true)
		timers[k] = nil
	end
end

function nanoTalk:OnProfileChanged(event, database, newProfileKey)
	db = database.profile
end

function nanoTalk:CHAT_MSG_WHISPER(event, msg, player)
	local now = GetTime()
	local last

	for tm, person in pairs(people) do
		if person.name == player then
			last = tm
			break
		end
	end
	local msgcnt = 0

	if last then
		msgcnt = people[last].msgs
		people[last] = nil

		for idx, tm in ipairs(timestamps) do
			if tm == last then
				tremove(timestamps, idx)
				break
			end
		end
	end
	people[now] = {
		name = player,
		msgs = msgcnt + 1
	}
	tinsert(timestamps, now)

	if #timestamps > 1 then
		tsort(timestamps)
	end

	if db.SoundAlert then
		PlaySoundFile("Interface\\AddOns\\nanoTalk\\beep.wav")
	end

	if db.NotifyAFK and UnitIsAFK("player") then
		SendChatMessage("I have been AFK for "..TimeStr(now - AFKtime)..".", "WHISPER", nil, player)
	elseif not last or ((now - last) > talkSecs) then
		local count = GetCount(talkSecs) - 1

		if count < 1 then
			UpdateTooltip()
			return
		end

		if db.ChatNotify then
			local phrase = (count > 1) and string.format("I have %d other people", count) or "Someone else is"

			SendChatMessage(string.format("Hello, %s! %s talking to me, so I might not answer right away.", player, phrase), "WHISPER", nil, player)
			self:Print(ChatFrame1, string.format("Informed %s of conversation.", player))
		end
	end
	UpdateTooltip()
end

function nanoTalk:UpdateInfo()
	if #timestamps > db.TrackNum then
		local ppos = tremove(timestamps)
		people[ppos] = nil
	end
	local now = GetTime()
	local is_afk = UnitIsAFK("player")

	if is_afk and AFKtime == 0 then
		AFKtime = now
	elseif not is_afk and AFKtime > 0 then
		self:Print(ChatFrame1, "You were AFK for "..TimeStr(now - AFKtime)..".")
		AFKtime = 0
	end
	local total = GetCount(trackSecs)
	local recent = GetCount(talkSecs)

	if recent > 0 then
		DataObj.text = ""..total.."t "..recent.."r"
		DataObj.icon = "Interface\\GossipFrame\\PetitionGossipIcon"
	else
		DataObj.text = ""..total.."t "
		DataObj.icon = "Interface\\GossipFrame\\GossipGossipIcon"
	end

	if tooltip and tooltip:IsShown() and #updatedLines then
		for linenum, tm in pairs(updatedLines) do
			tooltip:SetCell(linenum, 3, TimeStr(now - tm))
		end
	end
end

function nanoTalk:Clear()
	wipe(timestamps)
	wipe(people)
	UpdateTooltip()
end

-------------------------------------------------------------------------------
-- Configuration
-------------------------------------------------------------------------------
local ACD = LibStub("AceConfigDialog-3.0")

local function GetOptions()
	if not nanoTalk.options then
		nanoTalk.options = {
			type = "group",
			name = "nanoTalk",
			args = {
				talktime = {
					type = "range",
					name = "Conversation Timer",
					desc = "How many minutes you will be considered to be in a conversation.",
					min = 1,
					max = 10,
					step = 1,
					order = 100,
					get = function() return db.TalkTime end,
					set = function(info, v)
						      db.TalkTime = v
						      SetTimes()
					      end
				},
				tracktime = {
					type = "range",
					name = "Tracking Timer",
					desc = "How many minutes to track people who have sent you a message. (Set to zero to track indefinitely)",
					min = 0,
					max = 120,
					step = 1,
					order = 110,
					get = function() return db.TrackTime end,
					set = function(info, v)
						      db.TrackTime = v
						      SetTimes()
					      end
				},
				tracknum = {
					type = "range",
					name = "People to Track",
					desc = "Maximum number of people to track.",
					min = 0,
					max = 50,
					step = 1,
					order = 120,
					get = function() return db.TrackNum end,
					set = function(info, v) db.TrackNum = v end
				},
				tipscale = {
					type = "range",
					name = "Tooltip Scale",
					desc = "Size of tooltip and its text.",
					min = 0.5,
					max = 1.5,
					step = 0.05,
					order = 121,
					get = function() return db.TipScale end,
					set = function(info, v) db.TipScale = v end
				},
				null_1 = {
					order = 122,
					type = "description",
					name = " ",
					cmdHidden = true
				},
				afknotify = {
					type = "toggle",
					name = "AFK Notification",
					desc = "Notify others when you are AFK.",
					order = 130,
					get = function() return db.NotifyAFK end,
					set = function(info, v) db.NotifyAFK = v end
				},
				chatnotify = {
					type = "toggle",
					name = "Chat Notification",
					desc = "Notify others when you are in a conversation.",
					order = 140,
					get = function() return db.ChatNotify end,
					set = function(info, v) db.ChatNotify = v end
				},
				soundalert = {
					type = "toggle",
					name = "Sound Alert",
					desc = "Beep when receiving a message.",
					order = 150,
					get = function() return db.SoundAlert end,
					set = function(info, v) db.SoundAlert = v end
				},
			},
		}
	end
	return nanoTalk.options
end


function nanoTalk:SetupOptions()
	if not self.optionsFrames then self.optionsFrames = {} end

	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("nanoTalk", GetOptions())
	LibStub("AceConfig-3.0"):RegisterOptionsTable("nanoTalk", GetOptions(), "nanotalk")

	self.optionsFrames.nanoTalk = ACD:AddToBlizOptions("nanoTalk")
	LibStub("tekKonfig-AboutPanel").new("nanoTalk", "nanoTalk")
end
